#!/usr/bin/env python
import os

import numpy as np
from shapely.geometry import LineString, MultiPoint, Point
from shapely.ops import unary_union


class Interactor:
    def __init__(
        self,
        name="",
        frame=-1,
        nodes=[],
        coords_array=np.array([]),
        skeleton={},
        model={},
        interactions={},
        **kwargs,
    ):

        # Assign the name and frame of the interactor
        self.name = name
        self.frame = frame

        # Assign the nodes, require to assign the coords_array
        self._nodes = nodes

        # Create a dicts to store the skeleton, models, and interactions
        self._skeleton = {}
        self._models = {}
        self._interactions = interactions

        # Update the skeleton using the coords_array and model
        for _sk in skeleton:
            self._updateSkeleton(coords_array, **_sk)

        # Update the skeleton using the interactions using the skeleton
        for _m in model:
            self._updateModel(**_m)

    def _updateSkeleton(
        self, coords_array, name="", shape="", nodes=[], buffer=0, exclude_nodes=[]
    ):

        # Check if shape has all nodes
        if exclude_nodes:
            nodes = [_n for _n in self._nodes if _n not in exclude_nodes]

        # Confirm a name was given for the skeleton
        if not name:
            raise Exception(
                f"No name given for skeleton with shape ({shape}) and nodes ({', '.join(nodes)})"
            )

        # Check the shape is allowed
        if shape not in ["line", "reduced-line", "point", "hull"]:
            raise Exception(f"The shape ({shape}) is not supported")

        # Confirm the nodes are supported
        if shape in ["line", "reduced-line"] and len(nodes) <= 1:
            raise Exception(
                f"The shape ({shape}) requires two or more nodes, found {len(nodes)} nodes."
            )
        elif shape == "hull" and len(nodes) <= 1:
            raise Exception(
                f"The shape ({shape}) requires two or more nodes, found {len(nodes)} nodes."
            )
        elif shape == "point" and len(nodes) != 1:
            raise Exception(
                f"The shape ({shape}) requires a single node, found {len(nodes)} nodes."
            )

        # Assign values
        node_coords = [coords_array[self._nodes.index(_n)] for _n in nodes]

        # Update the skeleton w/ the correct shape and values
        if shape == "line":
            self._skeleton[name] = LineString(node_coords).buffer(buffer)
        if shape == "reduced-line":
            line_string = LineString(node_coords[:2])
            node_coords[0] = line_string.interpolate(buffer)
            if len(node_coords) == 2:
                node_coords[-1] = line_string.interpolate(line_string.length - buffer)
                self._skeleton[name] = LineString(node_coords).buffer(buffer)
            else:
                end_line_string = LineString(node_coords[-2:])
                node_coords[-1] = end_line_string.interpolate(
                    end_line_string.length - buffer
                )
                self._skeleton[name] = LineString(node_coords).buffer(buffer)
        elif shape == "hull":
            self._skeleton[name] = MultiPoint(node_coords).convex_hull.buffer(buffer)
        elif shape == "point":
            self._skeleton[name] = Point(node_coords[0]).buffer(buffer)

    def _updateModel(self, name="", includes=[], excludes=[]):

        # Confirm a name was given for the interaction
        if not name:
            raise Exception(
                f"No name given for interaction model with includes ({', '.join(includes)}) and excludes ({', '.join(excludes)})"
            )

        # Create the interaction with the inclusion skeleton nodes
        interaction = unary_union([self._skeleton[_ic] for _ic in includes])

        # Check if there are exclusion skeleton nodes
        if len(excludes):

            # Create the interaction to exclude
            exclude_interaction = unary_union([self._skeleton[_ic] for _ic in excludes])

            # Create the interaction without the exclusion
            interaction = interaction.difference(exclude_interaction)

        # Update the interaction
        self._models[name] = interaction

    def interacts(self, other_interactor):
        def _processInteraction(
            name="", origin="", dest="", symmetical=False, return_int=False
        ):
            def _interact(origin_interactor, dest_interactor):

                # Check if two interact
                interact = origin_interactor._models[origin].intersects(
                    dest_interactor._models[dest]
                )
                return interact

            def _reportInteract(interact, origin_interactor, dest_interactor):

                # Return int if specified
                if return_int:
                    interact = int(interact)

                # Return the interaction
                return {
                    "Origin interactor": origin_interactor.name,
                    "Destination interactor": dest_interactor.name,
                    "Interaction Name": name,
                    "Interaction Frame": self.frame,
                }

            # Confirm a name was given for the interaction
            if not name:
                raise Exception(
                    f"No name given for interaction with origin ({origin}) and dest ({dest})"
                )

            # Confirm a origin and dest were given
            if not origin:
                raise Exception(f"No origin given for interaction ({name})")
            if not dest:
                raise Exception(f"No dest given for interaction ({name})")

            # Confirm the origin and dest are in models
            if origin not in self._models:
                raise Exception(
                    f"Given origin ({origin}) not found among interaction models"
                )
            if dest not in self._models:
                raise Exception(
                    f"Given dest ({dest}) not found among interaction models"
                )

            # Confirm the type of symmetical
            if not isinstance(symmetical, bool):
                raise Exception(f"symmetical must be boolean")

            # Report symmetical results
            if symmetical:
                interaction_bool = _interact(self, other_interactor)
                if not interaction_bool:
                    return []
                interaction = _reportInteract(interaction_bool, self, other_interactor)
                return [interaction]

            # Report nonsymmetical results
            return_interactions = []

            # Test for an interaction
            interaction_bool = _interact(self, other_interactor)
            if interaction_bool:
                return_interactions.append(
                    _reportInteract(interaction_bool, self, other_interactor)
                )

            # Test for an inverse interaction
            inverse_interaction_bool = _interact(other_interactor, self)
            if inverse_interaction_bool:
                return_interactions.append(
                    _reportInteract(inverse_interaction_bool, other_interactor, self)
                )

            return return_interactions

        # Create a list to store the results
        interaction_list = []

        # Iterate the interaction
        for interaction in self._interactions:
            interaction_list.extend(_processInteraction(**interaction))

        return interaction_list

    @classmethod
    def withModelDict(cls, name, frame, nodes, coords_array, model_dict, **kwargs):
        return cls(
            name=name,
            frame=frame,
            nodes=nodes,
            coords_array=coords_array,
            skeleton=model_dict["Skeleton"],
            model=model_dict["Models"],
            interactions=model_dict["Interactions"],
            **kwargs,
        )



#!/usr/bin/env python
import csv
from itertools import combinations
from multiprocessing import Pool

import cv2
import matplotlib
import numpy as np
import pandas as pd
import sleap
import yaml

from naps.sleap_utils import load_tracks_from_slp
from naps.utils.interactor import Interactor


class TrackingArray:
    def __init__(
        self,
        frames_coords_array="",
        tracks=[],
        nodes=[],
        interactor_model_dict={},
        chunk_size=5000,
        threads=1,
        **kwargs,
    ):

        # Assign the frame coordsinates array
        self._frames_coords_array = frames_coords_array

        # Assign the tracks and nodes
        self._tracks = tracks
        self._nodes = nodes

        # Store the Interactor model
        self._interactor_model_dict = interactor_model_dict

        # Assign the frame chunk size
        self._chunk_size = chunk_size

        # Assign empty variables for processing
        self._process_frame_start = 0
        self._process_frame_end = 0
        self._process_coords_array = None

        # Assign general variables
        self._threads = threads

    def assign(self, out_file, frame_start=0, frame_end=-1):

        # Update end frame
        frame_end = frame_end if frame_end != -1 else self._frames_coords_array.shape[0]

        # Assign the pair-wise combinations of each track
        def _yield_combinations():
            for _tA, _tB in combinations(range(len(self._tracks)), 2):
                yield _tA, _tB

        with open(out_file, "w") as outfile:
            csv_writer = csv.DictWriter(
                outfile,
                fieldnames=[
                    "Origin interactor",
                    "Destination interactor",
                    "Interaction Name",
                    "Interaction Frame",
                ],
                delimiter="\t",
            )
            csv_writer.writeheader()

            for chunk_frame_start in range(frame_start, frame_end, self._chunk_size):

                # Store the frame start for the current processing chunk
                self._process_frame_start = chunk_frame_start

                # Store the frame end for the current processing chunk
                chunk_end = chunk_frame_start + self._chunk_size
                self._process_frame_end = (
                    frame_end if frame_end < chunk_end else chunk_end
                )

                # Store the processing chunked coordsinates array
                self._process_coords_array = self._frames_coords_array[
                    self._process_frame_start : self._process_frame_end, :, :, :
                ]

                if self._threads > 1:
                    with Pool(processes=self._threads) as pool:
                        for report_list in pool.starmap(
                            self._checkForInteractions, _yield_combinations()
                        ):
                            for report in report_list:
                                csv_writer.writerow(report)
                else:
                    for tagA, tagB in _yield_combinations():
                        for report in self._checkForInteractions(tagA, tagB):
                            csv_writer.writerow(report)

    def plot(
        self,
        interactions_filename,
        models_to_plot,
        video_filename,
        out_file,
        frame_start=0,
        frame_end=-1,
        limit_interactions=[],
        plot_color=None,
        line_width=2,
    ):

        # Convert to int32 for cv2 coords
        def int_coords(coords):
            return np.array(coords).round().astype(np.int32)

        # Update end frame
        frame_end = frame_end if frame_end != -1 else self._frames_coords_array.shape[0]
        frame_end = (
            frame_end - 1
            if frame_end >= self._frames_coords_array.shape[0]
            else frame_end
        )

        # Create dataframe of instances to plot per frame
        frame_instances_dataframe = self.processInteractionsFile(
            interactions_filename,
            frame_start,
            frame_end,
            limit_interactions=limit_interactions,
        )

        # Assign color for plotting the model:
        # 1) By default, plot red
        # 2) If given a list or array, check and assign the color
        # 3) If given a str, assign the color from matplotlib
        if plot_color == None:
            plot_color = (255, 0, 0)
        elif isinstance(plot_color, (list, tuple)):
            if len(plot_color) != 3:
                raise Exception(
                    f"Unable to assign plot color: {plot_color}. Iterables must be of length 3"
                )
            plot_color = tuple(plot_color)
        elif isinstance(plot_color, str):
            plot_color = matplotlib.colors.to_rgb(plot_color)
            plot_color = tuple(_v * 255 for _v in plot_color)

        # Color the RGB to BGR used by CV2
        plot_color = plot_color[::-1]

        # Set the current frame of the job
        current_frame = frame_start

        # Initialize OpenCV, then set the starting frame (0-based)
        video = cv2.VideoCapture(video_filename)
        video.set(cv2.CAP_PROP_POS_FRAMES, current_frame)

        # Store the width, height, and fps of the input video
        video_width = int(video.get(cv2.CAP_PROP_FRAME_WIDTH))
        video_height = int(video.get(cv2.CAP_PROP_FRAME_HEIGHT))
        video_fps = int(video.get(cv2.CAP_PROP_FPS))

        # Open the output video using the input video values
        out_video = cv2.VideoWriter(
            out_file,
            cv2.VideoWriter_fourcc(*"mp4v"),
            video_fps,
            (video_width, video_height),
        )

        # Read in the frame, confirm it was successful
        frame_read, frame_img = video.read()
        while frame_read and current_frame <= frame_end:

            # Assign the instances to plot
            instances_to_plot = frame_instances_dataframe[
                frame_instances_dataframe["Interaction Frame"] == current_frame
            ]["Instances"].values
            if len(instances_to_plot) > 1:
                raise Exception("Error parsing interactions for instances")
            if len(instances_to_plot) > 0:
                instances_to_plot = instances_to_plot[0]

            # Loop each instance for this frame
            for instance_to_plot in instances_to_plot:

                # Create an interactor using the instance
                instance_interactor = Interactor.withModelDict(
                    instance_to_plot,
                    current_frame,
                    self._nodes,
                    self._frames_coords_array[
                        current_frame, :, :, self._tracks.index(instance_to_plot)
                    ],
                    self._interactor_model_dict,
                )

                # Loop and plot each model for the instance
                for model_to_plot in models_to_plot:
                    instance_model_coords = int_coords(
                        instance_interactor._models["Head_w_antennae"].exterior.coords
                    )
                    cv2.polylines(
                        frame_img, [instance_model_coords], True, plot_color, line_width
                    )

            # Write the frame
            out_video.write(frame_img)

            # Read the next frame and advance the current frame
            frame_read, frame_img = video.read()
            current_frame += 1

        # Release the video
        out_video.release()

    def _checkForInteractions(self, trackA, trackB):

        # Create list to store the interactions
        interaction_list = []

        # Subset the array by tracks
        filtered_coords_array = np.take(self._process_coords_array, [trackA, trackB], 3)

        # Check each row for an interaction
        for frame, row in enumerate(filtered_coords_array):

            # Skip if missing data
            if np.isnan(row).any():
                continue

            # Adjust the frame to account for the start frame
            adj_frame = self._process_frame_start + frame

            # Create the interactors
            interactor1 = Interactor.withModelDict(
                self._tracks[trackA],
                adj_frame,
                self._nodes,
                row[:, :, 0],
                self._interactor_model_dict,
            )
            interactor2 = Interactor.withModelDict(
                self._tracks[trackB],
                adj_frame,
                self._nodes,
                row[:, :, 1],
                self._interactor_model_dict,
            )

            # Check if there is an interaction
            interaction_list.extend(interactor1.interacts(interactor2))

        return interaction_list

    @classmethod
    def fromSLEAP(cls, sleap_file, model_dict={}, model_yaml="", **kwargs):

        # Assign the tracks from the input file
        labels = sleap.Labels.load_file(sleap_file)
        tracks = [_k.name for _k in labels.tracks]
        labels = None

        # Create an array of [frames, nodes, (x,y), tracks] from the input file
        frames_coords_array, nodes = load_tracks_from_slp(sleap_file)

        # Confirm one a single model was given
        if model_dict and model_yaml:
            raise Exception("Only a single Interactor model may be specified")

        # Assign the model dict, if specified
        if model_yaml:
            interactor_model_dict = TrackingArray.loadYAML(model_yaml)
        elif model_dict:
            interactor_model_dict = model_dict

        return cls(
            frames_coords_array=frames_coords_array,
            tracks=tracks,
            nodes=nodes,
            interactor_model_dict=interactor_model_dict,
            **kwargs,
        )

    @staticmethod
    def loadYAML(yaml_filename):

        # Read in the YAML file and return as dict
        with open(yaml_filename, "r") as yaml_file:
            try:
                yaml_dict = yaml.safe_load(yaml_file)
            except yaml.YAMLError as err:
                raise Exception(err)
        return yaml_dict

    @staticmethod
    def processInteractionsFile(
        interactions_file,
        frame_start,
        frame_end,
        limit_interactions=[],
        expected_cols=[
            "Origin interactor",
            "Destination interactor",
            "Interaction Name",
            "Interaction Frame",
        ],
    ):

        # Process the interactions file
        interactions_origin_dataframe = pd.read_csv(interactions_file, sep="\t")[
            expected_cols
        ]
        if limit_interactions:
            interactions_origin_dataframe = interactions_origin_dataframe[
                interactions_origin_dataframe["Interaction Name"].isin(
                    limit_interactions
                )
            ]
        interactions_dest_dataframe = interactions_origin_dataframe.copy()

        # Assign the origin instances for each frame
        interactions_origin_dataframe = interactions_origin_dataframe.drop(
            columns=["Destination interactor"]
        )
        interactions_origin_dataframe = interactions_origin_dataframe.rename(
            columns={"Origin interactor": "Instance"}
        )

        # Assign the destination instances for each frame
        interactions_dest_dataframe = interactions_dest_dataframe.drop(
            columns=["Origin interactor"]
        )
        interactions_dest_dataframe = interactions_dest_dataframe.rename(
            columns={"Destination interactor": "Instance"}
        )

        # Create the instance dataframe
        instance_dataframe = interactions_origin_dataframe.append(
            interactions_dest_dataframe, ignore_index=True
        )
        instance_dataframe = instance_dataframe[
            instance_dataframe["Interaction Frame"].between(frame_start, frame_end)
        ]

        # Check the presence of interactions within the given frames
        if instance_dataframe.shape[0] == 0:
            raise Exception(
                f"No interactions found between frames {frame_start} to {frame_end}"
            )

        instance_dataframe = (
            instance_dataframe.groupby("Interaction Frame")["Instance"]
            .apply(list)
            .reset_index(name="Instances")
        )

        return instance_dataframe

    @staticmethod
    def interpolate(coords_array):
        import pandas as pd

        for i in range(coords_array.shape[3]):

            coords_df = pd.DataFrame(coords_array[:, :, 0, i])
            coords_df = coords_df.interpolate(method="linear", axis=0)
            diff_df = coords_df.diff().abs().ge(5)
            coords_df[diff_df] = np.nan
            coords_array[:, :, 0, i] = coords_df.to_numpy()

            coords_df = pd.DataFrame(coords_array[:, :, 1, i])
            coords_df = coords_df.interpolate(method="linear", axis=0)
            diff_df = coords_df.diff().abs().ge(5)
            coords_df[diff_df] = np.nan
            coords_array[:, :, 1, i] = coords_df.to_numpy()

        return coords_array



#!/usr/bin/env python
import argparse
import logging
import os
import time

from naps.utils.tracking import TrackingArray

logger = logging.getLogger("NAPS Logger")


def plotParser():
    """
    Plot Parser

    Assign the parameters for plot assignment

    Parameters
    ----------
    sys.argv : list
            Parameters from command line

    Raises
    ------
    IOError
            If the specified files do not exist
    """

    def confirmFile():
        """Custom action to confirm file exists"""

        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError("%s not found" % value)
                setattr(args, self.dest, value)

        return customAction

    plot_parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Input files
    plot_parser.add_argument(
        "--sleap-file",
        help="Defines the input SLEAP file",
        type=str,
        action=confirmFile(),
        required=True,
    )
    plot_parser.add_argument(
        "--yaml-model",
        help="Defines the YAML model file",
        type=str,
        action=confirmFile(),
        required=True,
    )

    plot_parser.add_argument(
        "--interactions-file",
        help="Defines the interactions file",
        type=str,
        required=True,
    )
    plot_parser.add_argument(
        "--limit-interactions",
        help="Limits the interactions to plot",
        type=str,
        nargs="+",
    )

    plot_parser.add_argument(
        "--video-file", help="Defines the input video file", type=str, required=True
    )
    plot_parser.add_argument(
        "--models-to-plot",
        help="Defines the models to plot",
        type=str,
        nargs="+",
        required=True,
    )

    # General options
    plot_parser.add_argument(
        "--frame-start",
        help="Defines the frame to start interaction assignment",
        type=int,
        default=0,
    )
    plot_parser.add_argument(
        "--frame-end",
        help="Defines the frame to end interaction assignment",
        type=int,
        default=-1,
    )
    plot_parser.add_argument(
        "--plot-color", help="Defines the color of the models", type=str, default="red"
    )
    plot_parser.add_argument(
        "--line-width", help="Defines the line width of the models", type=int, default=2
    )

    # Output options
    plot_parser.add_argument(
        "--out-file", help="Defines the output video file", type=str, required=True
    )

    # Return the arguments
    return plot_parser.parse_args()


def main():

    # Assign the interaction args
    plot_args = plotParser()

    # Load the tracking array
    logger.info("Loading the tracking array...")
    t0 = time.time()

    # Create the tracking array object using a SLEAP file and the yaml model
    tracking_array = TrackingArray.fromSLEAP(
        plot_args.sleap_file, model_yaml=plot_args.yaml_model
    )

    logger.info("Done loading in %s seconds.", time.time() - t0)

    # Plot interactions from the tracking array
    logger.info("Starting plotting...")
    logger.info(f"Ploting frames {plot_args.frame_start} to {plot_args.frame_end}")
    t0 = time.time()

    tracking_array.plot(
        plot_args.interactions_file,
        plot_args.models_to_plot,
        plot_args.video_file,
        plot_args.out_file,
        frame_start=plot_args.frame_start,
        frame_end=plot_args.frame_end,
        limit_interactions=plot_args.limit_interactions,
    )

    logger.info("Done plotting in %s seconds.", time.time() - t0)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
import logging
import time
from collections import defaultdict

import sleap

from naps.aruco import ArUcoModel
from naps.matching import Matching
from naps.sleap_utils import update_labeled_frames

logger = logging.getLogger("NAPS Logger")


def build_parser():
    parser = argparse.ArgumentParser(
        description="NAPS -- Hybrid tracking using SLEAP and ArUco tags"
    )

    parser.add_argument(
        "--slp-path",
        help="The filepath of the SLEAP (.slp or .h5) file to pull coordinates from. This should correspond with the input video file. The SLEAP file contains pose estimation data required for NAPS.",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--h5-path",
        help="The filepath of the analysis h5 file to pull coordinates from. This should correspond with the input video file and slp file. Because NAPS relies on having data in the format given by SLEAP analysis h5s, you can skip a step by providing the analysis h5 file directly. If this is not provided, NAPS will pull the data from the SLEAP file.",
        type=str,
        default=None,
    )

    parser.add_argument(
        "--video-path",
        help="The filepath of the video used with SLEAP. This should be the video file that SLEAP was run on.",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--tag-node-name",
        help="The ArUco tag SLEAP node name. This should correspond to the node name of the ArUco tag in the SLEAP data. 'tag' is set as the default value.",
        type=str,
        default="tag",
    )

    parser.add_argument(
        "--start-frame",
        help="The zero-based fully-closed frame to begin NAPS assignment. This value allows you to specify from which frame to start the processing.",
        type=int,
        required=True,
    )

    parser.add_argument(
        "--end-frame",
        help="The zero-based fully-closed frame to stop NAPS assignment. This allows you to specify at which frame to end the processing.",
        type=int,
        required=True,
    )

    parser.add_argument(
        "--half-rolling-window-size",
        help="Specifies the number of flanking frames (prior and subsequent) required in the rolling window for Hungarian matching a frame. The larger this window, the more frames will be used for matching. This can result in more robust tracks but in the event of an identity swap, it may take longer to correct the identity. The default is set to 5, resulting in an 11 frame window. You should change this depending on your intended use and your recording setup. ",
        type=int,
        default=5,
    )

    parser.add_argument(
        "--output-path",
        help="Output path of the resulting SLEAP h5 analysis file. The default value is 'naps_output.analysis.h5', storing the output in the current working directory. If a different location or file name is preferred, it can be specified here.",
        type=str,
        default="naps_output.analysis.h5",
    )

    parser.add_argument(
        "--aruco-crop-size",
        help="The number of pixels horizontally and vertically around the ArUco SLEAP node to identify the marker. The cropping area should be large enough to include the whole marker for accurate detection, but not much larger, to keep the processing efficient and avoid capturing multiple tags. This is data set specific.",
        type=int,
    )

    parser.add_argument(
        "--aruco-marker-set",
        help="The ArUco markers used in the video. This must match the specific set of ArUco markers used in the video for accurate detection and tracking. An example would be DICT_5X5_50, which is a set of 50 5x5 markers. More information can be found in the OpenCV documentation.",
        type=str,
        required=True,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-win-size-min",
        dest="adaptiveThreshWinSizeMin",
        help="Specifies the value for adaptiveThreshWinSizeMin used in adaptive thresholding. This parameter affects the adaptive thresholding in the ArUco marker detection, which can impact the robustness of marker detection. The default value is 10, a commonly used value. More information can be found in the OpenCV documentation.",
        type=int,
        default=10,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-win-size-max",
        dest="adaptiveThreshWinSizeMax",
        help="Specifies the value for adaptiveThreshWinSizeMax used in adaptive thresholding. This parameter, similar to adaptiveThreshWinSizeMin, influences the adaptive thresholding. The default value is 30, providing a larger window size for the thresholding process, improving marker detection under diverse lighting conditions. More information can be found in the OpenCV documentation.",
        type=int,
        default=30,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-win-size-step",
        dest="adaptiveThreshWinSizeStep",
        help="Specifies the value for adaptiveThreshWinSizeStep used in adaptive thresholding. This parameter determines the step size for the window in adaptive thresholding, affecting the granularity of the process. The default value is 12, offering a balanced choice between processing speed and thresholding precision. More information can be found in the OpenCV documentation.",
        type=int,
        default=12,
    )

    parser.add_argument(
        "--aruco-adaptive-thresh-constant",
        dest="adaptiveThreshConstant",
        help="Specifies the value for adaptiveThreshConstant used in adaptive thresholding. This parameter is a constant subtracted from the mean or weighted sum of the neighbourhood pixels. The default value is 3, which works well in most scenarios, but can be adjusted based on specific lighting conditions. More information can be found in the OpenCV documentation.",
        type=float,
        default=3,
    )

    parser.add_argument(
        "--aruco-perspective-rm-ignored-margin",
        dest="perspectiveRemoveIgnoredMarginPerCell",
        help="Specifies the value for perspectiveRemoveIgnoredMarginPerCell. This parameter is used in the perspective removal of the marker. The default value is 0.1, which is a reasonable value for many situations. More information can be found in the OpenCV documentation.",
        type=float,
        default=0.1,
    )

    parser.add_argument(
        "--aruco-error-correction-rate",
        dest="errorCorrectionRate",
        help="Specifies the value for errorCorrectionRate. This parameter is used for error correction when decoding ArUco tags. The default value is 1 is to fully utilize the error capability for each tag dictionary. Adjust this parameter can be necessary for improving the robustness of marker detection, especially in the presence of camera noise or occlusions. More information can be found in the OpenCV documentation.",
        type=float,
        default=1,
    )

    return parser


def main(argv=None):
    """Main function for the NAPS tracking script."""

    # Set the start time for the entire pipeline
    t0_total = time.time()

    # Build the arguments from the .parse_args(args)
    args = build_parser().parse_args(argv)

    # Assign the h5 path if not specified
    if args.h5_path is None:
        args.h5_path = args.slp_path

    # Create a track array from the SLEAP file(s)
    logger.info("Loading predictions...")
    t0 = time.time()
    # locations, node_names = load_tracks_from_slp(args.h5_path)
    tag_locations_dict = defaultdict(lambda: defaultdict(tuple))
    labels = sleap.Labels.load_file(args.slp_path)
    for lf in labels.labeled_frames:
        if lf.frame_idx < args.start_frame or lf.frame_idx > args.end_frame:
            continue
        for instance in lf.instances:
            tag_idx = instance.skeleton.node_names.index(args.tag_node_name)
            track_name = int(instance.track.name.split("_")[-1])
            # print(f'Frame {lf.frame_idx} track {track_name} has points {instance.numpy()[tag_idx]}')
            tag_locations_dict[lf.frame_idx][track_name] = instance.numpy()[tag_idx]

    # Create an ArUcoModel with the default/specified parameters
    logger.info("Create ArUco model...")
    t0 = time.time()
    aruco_model = ArUcoModel.withTagSet(
        args.aruco_marker_set,
        adaptiveThreshWinSizeMin=args.adaptiveThreshWinSizeMin,
        adaptiveThreshWinSizeMax=args.adaptiveThreshWinSizeMax,
        adaptiveThreshWinSizeStep=args.adaptiveThreshWinSizeStep,
        adaptiveThreshConstant=args.adaptiveThreshConstant,
        perspectiveRemoveIgnoredMarginPerCell=args.perspectiveRemoveIgnoredMarginPerCell,
        errorCorrectionRate=args.errorCorrectionRate,
    )
    logger.info("ArUco model built in %s seconds.", time.time() - t0)

    # Match the track to the ArUco markers
    logger.info("Starting matching...")
    t0 = time.time()
    matching = Matching(
        args.video_path,
        args.start_frame,
        args.end_frame,
        marker_detector=aruco_model.detect,
        aruco_crop_size=args.aruco_crop_size,
        half_rolling_window_size=args.half_rolling_window_size,
        tag_node_dict=tag_locations_dict,
    )
    matching_dict = matching.match()
    logger.info("Done matching in %s seconds.", time.time() - t0)

    # Create the output
    logger.info("Reconstructing SLEAP file...")
    t0 = time.time()
    # Right now the reconstruction assumes that we each track has a single track ID assigned to it. We'll generalize so that a track can switch IDs over time.
    labels = update_labeled_frames(
        args.slp_path, matching_dict, args.start_frame, args.end_frame
    )
    labels.save(args.output_path)
    logger.info("Done reconstructing SLEAP file in %s seconds.", time.time() - t0)

    logger.info("Complete NAPS runtime: %s", time.time() - t0_total)


if __name__ == "__main__":
    main()



#!/usr/bin/env python
import argparse
import logging
import os
import time

from naps.utils.tracking import TrackingArray

logger = logging.getLogger("NAPS Logger")


def interactionParser():
    """
    Interaction Parser

    Assign the parameters for interaction assignment

    Parameters
    ----------
    sys.argv : list
            Parameters from command line

    Raises
    ------
    IOError
            If the specified files do not exist
    """

    def confirmFile():
        """Custom action to confirm file exists"""

        class customAction(argparse.Action):
            def __call__(self, parser, args, value, option_string=None):
                if not os.path.isfile(value):
                    raise IOError("%s not found" % value)
                setattr(args, self.dest, value)

        return customAction

    interaction_parser = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter
    )

    # Input files
    interaction_parser.add_argument(
        "--sleap-file",
        help="Defines the input SLEAP file",
        type=str,
        action=confirmFile(),
        required=True,
    )
    interaction_parser.add_argument(
        "--yaml-model",
        help="Defines the YAML model file",
        type=str,
        action=confirmFile(),
        required=True,
    )

    # General options
    interaction_parser.add_argument(
        "--frame-start",
        help="Defines the frame to start interaction assignment",
        type=int,
        default=0,
    )
    interaction_parser.add_argument(
        "--frame-end",
        help="Defines the frame to end interaction assignment",
        type=int,
        default=-1,
    )
    interaction_parser.add_argument(
        "--threads", help="Number of threads to use", type=int, default=1
    )

    # interaction_parser.add_argument('--max-velocity', help = 'Defines the maximum velocity allowed in the dataset', type = int)
    # interaction_parser.add_argument('--remove-tags', help = 'Defines the tags to remove', type = str, nargs = '+', default = [])
    # interaction_parser.add_argument('--min-bin', help = 'Min interaction distance', type = int, default = 50)
    # interaction_parser.add_argument('--max-bin', help = 'Max interaction distance', type = int, default = 450)
    # interaction_parser.add_argument('--frame-buffer', help = 'Frame buffer - i.e. gap between frames allowed', type = int, default = 5)
    # interaction_parser.add_argument('--min-frames', help = 'Min frames for interaction', type = int, default = 5)

    # Output options
    interaction_parser.add_argument(
        "--out-file",
        help="Defines the output tsv file. Defaults to: out.tsv",
        type=str,
        default="out.tsv",
    )

    # Return the arguments
    return interaction_parser.parse_args()


def main():

    # Assign the interaction args
    interaction_args = interactionParser()

    # Load the tracking array
    logger.info("Loading the tracking array...")
    t0 = time.time()

    # Create the interactions object using a SLEAP file and the yaml model
    interactions = TrackingArray.fromSLEAP(
        interaction_args.sleap_file,
        model_yaml=interaction_args.yaml_model,
        threads=interaction_args.threads,
    )

    logger.info("Done loading in %s seconds.", time.time() - t0)

    # Assign interactions from the tracking array
    logger.info("Starting interaction assignment...")
    t0 = time.time()

    # Assign the interactions
    interactions.assign(
        interaction_args.out_file,
        frame_start=interaction_args.frame_start,
        frame_end=interaction_args.frame_end,
    )

    logger.info("Done with assignment in %s seconds.", time.time() - t0)


if __name__ == "__main__":
    main()
